#!/usr/bin/env node
'use strict';

function _interopDefault (ex) { return (ex && (typeof ex === 'object') && 'default' in ex) ? ex['default'] : ex; }

var puppeteer = _interopDefault(require('puppeteer'));
var url = require('url');
var chalk = _interopDefault(require('chalk'));
var prettyBytes = _interopDefault(require('pretty-bytes'));
var columnify = _interopDefault(require('columnify'));
var yargs = _interopDefault(require('yargs'));

let launch = (() => {
  var _ref = _asyncToGenerator(function* (opts = {}) {
    let browser;

    if (!browserWSEndpoint) {
      browser = yield puppeteer.launch(opts);
      browserWSEndpoint = browser.wsEndpoint();
    } else {
      browser = puppeteer.connect({ browserWSEndpoint });
    }

    return browser;
  });

  return function launch() {
    return _ref.apply(this, arguments);
  };
})();

function _asyncToGenerator(fn) { return function () { var gen = fn.apply(this, arguments); return new Promise(function (resolve, reject) { function step(key, arg) { try { var info = gen[key](arg); var value = info.value; } catch (error) { reject(error); return; } if (info.done) { resolve(value); } else { return Promise.resolve(value).then(function (value) { step("next", value); }, function (err) { step("throw", err); }); } } return step("next"); }); }; }

let browserWSEndpoint;

var _extends = Object.assign || function (target) { for (var i = 1; i < arguments.length; i++) { var source = arguments[i]; for (var key in source) { if (Object.prototype.hasOwnProperty.call(source, key)) { target[key] = source[key]; } } } return target; };

function _asyncToGenerator$1(fn) { return function () { var gen = fn.apply(this, arguments); return new Promise(function (resolve, reject) { function step(key, arg) { try { var info = gen[key](arg); var value = info.value; } catch (error) { reject(error); return; } if (info.done) { resolve(value); } else { return Promise.resolve(value).then(function (value) { step("next", value); }, function (err) { step("throw", err); }); } } return step("next"); }); }; }

let setPageOptions = (() => {
  var _ref = _asyncToGenerator$1(function* (page, options) {
    let { cookies, headers, userAgent, viewport } = options;

    if (typeof cookies !== 'undefined') {
      if (!Array.isArray(cookies)) {
        cookies = [cookies];
      }
      yield page.setCookie(...cookies);
    }

    if (typeof headers !== 'undefined') {
      yield page.setExtraHTTPHeaders(headers);
    }

    if (typeof userAgent !== 'undefined') {
      yield page.setUserAgent(userAgent);
    }

    if (typeof viewport !== 'undefined') {
      yield page.setViewport(viewport);
    }
  });

  return function setPageOptions(_x, _x2) {
    return _ref.apply(this, arguments);
  };
})();

var runner = (() => {
  var _ref2 = _asyncToGenerator$1(function* (_urls, options = {}) {
    let urls;
    if (Array.isArray(_urls)) {
      urls = _urls;
    } else {
      urls = [_urls];
    }

    urls = urls.map(function (url$$1) {
      let parsedUrl = url.parse(url$$1);
      if (!parsedUrl || !parsedUrl.protocol) {
        throw new Error('Url requires a valid protocol and host name.');
      }
      return parsedUrl;
    });

    const sourceURL = urls[0],
          { args = [], timeout } = options,
          browser = yield launch({ args: Array.isArray(args) ? args : [args], timeout }),
          page = yield browser.newPage(),
          { coverage } = page;

    // Set default options
    options = _extends({ js: true, css: true }, options);
    yield setPageOptions(page, options);

    // Map the requests to match covered requests later
    let requestMap = new Map();
    page.on('request', function (request) {
      let requestUrl = request.url(),
          req = url.parse(requestUrl);

      requestMap.set(requestUrl, _extends({
        type: request.resourceType(),
        headers: request.headers(),
        sameOrigin: !req.host ? true : req.host === sourceURL.host
      }, req));
    });

    yield Promise.all([options.js && coverage.startJSCoverage({ resetOnNavigation: false }), options.css && coverage.startCSSCoverage({ resetOnNavigation: false })]);

    try {
      for (let p of urls) {
        yield page.goto(p.href, { waitUntil: ['load', 'networkidle0'] });
      }
    } catch (ex) {
      yield browser.close();
      throw ex;
    }

    const [jsCoverage, cssCoverage] = yield Promise.all([options.js ? coverage.stopJSCoverage() : Promise.resolve([]), options.css ? coverage.stopCSSCoverage() : Promise.resolve([])]);

    let assetsMap = new Map(),
        entries = [...jsCoverage, ...cssCoverage];

    if (options.sameOrigin) {
      entries = entries.filter(function (entry) {
        return requestMap.get(entry.url).sameOrigin;
      });
    }

    for (const entry of entries) {
      if (!assetsMap.has(entry.url)) {
        let totalBytes = entry.text.length,
            usedBytes = 0,
            req = requestMap.get(entry.url) || {};

        for (const range of entry.ranges) {
          usedBytes += range.end - range.start;
        }

        assetsMap.set(entry.url, {
          path: req.sameOrigin ? req.path : req.href,
          type: req.type,
          totalBytes,
          usedBytes,
          unusedBytes: totalBytes - usedBytes,
          coverage: totalBytes !== usedBytes ? usedBytes / totalBytes * 100 : 100
        });
      }
    }

    yield browser.close();

    let assets = Array.from(assetsMap.values()),
        totalBytes = assets.reduce(function (totalBytes, assets) {
      return totalBytes + assets.totalBytes;
    }, 0),
        totalUsedBytes = assets.reduce(function (totalUsedBytes, assets) {
      return totalUsedBytes + assets.usedBytes;
    }, 0);

    return {
      url: sourceURL.href,
      host: sourceURL.host,
      totalBytes,
      totalUsedBytes,
      totalCoverage: totalUsedBytes / totalBytes * 100,
      coverages: assets
    };
  });

  return function (_x3) {
    return _ref2.apply(this, arguments);
  };
})();

function colorUsed(usedBytes, totalBytes) {
  let used = prettyBytes(usedBytes);

  if (totalBytes && usedBytes === 0) {
    used = chalk.red(used);
  } else {
    used = chalk.gray(used);
  }

  return used;
}

function colorUnused(unusedBytes, totalBytes) {
  let unused = prettyBytes(unusedBytes);

  if (unusedBytes && unusedBytes === totalBytes) {
    unused = chalk.red(unused);
  } else if (unusedBytes > 0) {
    unused = chalk.yellow(unused);
  } else {
    unused = chalk.green(unused);
  }

  return unused;
}

function colorCoverage(coveragePercent) {
  let precision = parseFloat(coveragePercent).toPrecision(coveragePercent < 10 ? 3 : 4),
      coverage = `${~~coveragePercent !== coveragePercent ? precision : coveragePercent}%`;

  if (coveragePercent === 100) {
    coverage = chalk.green(coverage);
  } else if (coveragePercent > 0) {
    coverage = chalk.yellow(coverage);
  } else {
    coverage = chalk.red(coverage);
  }

  return coverage;
}

function stats (coverage) {
  let { host, coverages } = coverage;

  coverages = coverages.map(entry => {
    let { path, type, totalBytes, usedBytes, unusedBytes, coverage: coveredPercent } = entry;

    const total = prettyBytes(totalBytes);
    const used = colorUsed(usedBytes, totalBytes);
    const unused = colorUnused(unusedBytes, totalBytes);
    const coverage = colorCoverage(coveredPercent);

    return { path, type, total, used, unused, coverage };
  });

  // Set right column alignments for numeric values
  const config = ['total', 'used', 'unused', 'coverage'].reduce((o, name) => (o[name] = { align: 'right' }) && o, {});

  const columns = columnify(coverages, { truncate: true, maxWidth: 120, minWidth: 8, config });

  return `${host} Coverage\n\n${columns}`;
}

function _asyncToGenerator$2(fn) { return function () { var gen = fn.apply(this, arguments); return new Promise(function (resolve, reject) { function step(key, arg) { try { var info = gen[key](arg); var value = info.value; } catch (error) { reject(error); return; } if (info.done) { resolve(value); } else { return Promise.resolve(value).then(function (value) { step("next", value); }, function (err) { step("throw", err); }); } } return step("next"); }); }; }

const BASIC_OPTIONS = 'Basic options:';
const PUPPETEER_OPTIONS = 'Puppeteer options:';

const _yargs = yargs.usage('\nUsage: $0 <url> [options]').options({

  // Basic Options
  'no-css': {
    type: 'boolean',
    describe: 'disables css code coverage',
    group: BASIC_OPTIONS
  },
  'no-js': {
    type: 'boolean',
    describe: 'disables js code coverage',
    group: BASIC_OPTIONS
  },
  origin: {
    type: 'boolean',
    alias: 'o',
    describe: 'only cover same-origin assets',
    group: BASIC_OPTIONS
  },

  // Puppeteer Options
  timeout: {
    type: 'number',
    alias: 't',
    describe: 'maximum number of milliseconds to wait for the browser to start',
    group: PUPPETEER_OPTIONS
  },
  viewportWidth: {
    type: 'number',
    alias: 'vw',
    describe: 'sets viewport width for page render',
    group: PUPPETEER_OPTIONS
  },
  viewportHeight: {
    type: 'number',
    alias: 'vh',
    describe: 'sets viewport height for page render',
    group: PUPPETEER_OPTIONS
  },
  viewport: {
    type: 'string',
    alias: 'V',
    describe: 'sets viewport for page render i.e. 800x600',
    group: PUPPETEER_OPTIONS
  },
  userAgent: {
    type: 'string',
    alias: 'ua',
    describe: 'sets the user agent for the request',
    group: PUPPETEER_OPTIONS
  },
  headers: {
    type: 'string',
    alias: 'H',
    describe: 'sets headers for the page request\nexample: $0 <url> -H "X-Header: 123"',
    group: PUPPETEER_OPTIONS
  }

}).array('headers').alias('help', 'h').alias('version', 'v').help();

function parseHeaders(headers) {
  return headers.reduce((headers, headerString) => {
    let [key, value] = headerString.split(/:\s+/);

    if (key && value) {
      headers[key] = value;
    }

    return headers;
  }, {});
}

_asyncToGenerator$2(function* () {
  let [url$$1] = _yargs.argv._,
      { origin, noCss, noJs, viewportHeight, viewportWidth, viewport, headers, userAgent } = yargs.argv;

  if (!url$$1) {
    yargs.showHelp();
    return;
  }

  if (viewport) {
    [viewportWidth, viewportHeight] = viewport.split('x').map(function (x) {
      return parseInt(x, 10) || void 0;
    });
  }

  if (typeof viewportWidth === 'undefined' && viewportHeight && typeof viewportHeight === 'number') {
    viewportWidth = ~~(viewportHeight * (4 / 3));
  }

  if (typeof viewportHeight === 'undefined' && viewportWidth && typeof viewportWidth === 'number') {
    viewportHeight = ~~(viewportHeight * (3 / 4));
  }

  if (viewportHeight && viewportWidth) {
    viewport = { height: viewportHeight, width: viewportWidth };
  }

  if (headers) {
    headers = parseHeaders(headers);
  }

  try {
    const coverage = yield runner(url$$1, { sameOrigin: origin, css: !noCss, js: !noJs, viewport, headers, userAgent });
    console.log(stats(coverage));
  } catch (err) {
    console.log(`${chalk.red('Error:')} ${err.message}`);
  }
})();
