import puppeteer from 'puppeteer';
import { parse } from 'url';

let launch = (() => {
  var _ref = _asyncToGenerator(function* (opts = {}) {
    let browser;

    if (!browserWSEndpoint) {
      browser = yield puppeteer.launch(opts);
      browserWSEndpoint = browser.wsEndpoint();
    } else {
      browser = puppeteer.connect({ browserWSEndpoint });
    }

    return browser;
  });

  return function launch() {
    return _ref.apply(this, arguments);
  };
})();

function _asyncToGenerator(fn) { return function () { var gen = fn.apply(this, arguments); return new Promise(function (resolve, reject) { function step(key, arg) { try { var info = gen[key](arg); var value = info.value; } catch (error) { reject(error); return; } if (info.done) { resolve(value); } else { return Promise.resolve(value).then(function (value) { step("next", value); }, function (err) { step("throw", err); }); } } return step("next"); }); }; }

let browserWSEndpoint;

var _extends = Object.assign || function (target) { for (var i = 1; i < arguments.length; i++) { var source = arguments[i]; for (var key in source) { if (Object.prototype.hasOwnProperty.call(source, key)) { target[key] = source[key]; } } } return target; };

function _asyncToGenerator$1(fn) { return function () { var gen = fn.apply(this, arguments); return new Promise(function (resolve, reject) { function step(key, arg) { try { var info = gen[key](arg); var value = info.value; } catch (error) { reject(error); return; } if (info.done) { resolve(value); } else { return Promise.resolve(value).then(function (value) { step("next", value); }, function (err) { step("throw", err); }); } } return step("next"); }); }; }

let setPageOptions = (() => {
  var _ref = _asyncToGenerator$1(function* (page, options) {
    let { cookies, headers, userAgent, viewport } = options;

    if (typeof cookies !== 'undefined') {
      if (!Array.isArray(cookies)) {
        cookies = [cookies];
      }
      yield page.setCookie(...cookies);
    }

    if (typeof headers !== 'undefined') {
      yield page.setExtraHTTPHeaders(headers);
    }

    if (typeof userAgent !== 'undefined') {
      yield page.setUserAgent(userAgent);
    }

    if (typeof viewport !== 'undefined') {
      yield page.setViewport(viewport);
    }
  });

  return function setPageOptions(_x, _x2) {
    return _ref.apply(this, arguments);
  };
})();

var runner = (() => {
  var _ref2 = _asyncToGenerator$1(function* (_urls, options = {}) {
    let urls;
    if (Array.isArray(_urls)) {
      urls = _urls;
    } else {
      urls = [_urls];
    }

    urls = urls.map(function (url) {
      let parsedUrl = parse(url);
      if (!parsedUrl || !parsedUrl.protocol) {
        throw new Error('Url requires a valid protocol and host name.');
      }
      return parsedUrl;
    });

    const sourceURL = urls[0],
          { args = [], timeout } = options,
          browser = yield launch({ args: Array.isArray(args) ? args : [args], timeout }),
          page = yield browser.newPage(),
          { coverage } = page;

    // Set default options
    options = _extends({ js: true, css: true }, options);
    yield setPageOptions(page, options);

    // Map the requests to match covered requests later
    let requestMap = new Map();
    page.on('request', function (request) {
      let requestUrl = request.url(),
          req = parse(requestUrl);

      requestMap.set(requestUrl, _extends({
        type: request.resourceType(),
        headers: request.headers(),
        sameOrigin: !req.host ? true : req.host === sourceURL.host
      }, req));
    });

    yield Promise.all([options.js && coverage.startJSCoverage({ resetOnNavigation: false }), options.css && coverage.startCSSCoverage({ resetOnNavigation: false })]);

    try {
      for (let p of urls) {
        yield page.goto(p.href, { waitUntil: ['load', 'networkidle0'] });
      }
    } catch (ex) {
      yield browser.close();
      throw ex;
    }

    const [jsCoverage, cssCoverage] = yield Promise.all([options.js ? coverage.stopJSCoverage() : Promise.resolve([]), options.css ? coverage.stopCSSCoverage() : Promise.resolve([])]);

    let assetsMap = new Map(),
        entries = [...jsCoverage, ...cssCoverage];

    if (options.sameOrigin) {
      entries = entries.filter(function (entry) {
        return requestMap.get(entry.url).sameOrigin;
      });
    }

    for (const entry of entries) {
      if (!assetsMap.has(entry.url)) {
        let totalBytes = entry.text.length,
            usedBytes = 0,
            req = requestMap.get(entry.url) || {};

        for (const range of entry.ranges) {
          usedBytes += range.end - range.start;
        }

        assetsMap.set(entry.url, {
          path: req.sameOrigin ? req.path : req.href,
          type: req.type,
          totalBytes,
          usedBytes,
          unusedBytes: totalBytes - usedBytes,
          coverage: totalBytes !== usedBytes ? usedBytes / totalBytes * 100 : 100
        });
      }
    }

    yield browser.close();

    let assets = Array.from(assetsMap.values()),
        totalBytes = assets.reduce(function (totalBytes, assets) {
      return totalBytes + assets.totalBytes;
    }, 0),
        totalUsedBytes = assets.reduce(function (totalUsedBytes, assets) {
      return totalUsedBytes + assets.usedBytes;
    }, 0);

    return {
      url: sourceURL.href,
      host: sourceURL.host,
      totalBytes,
      totalUsedBytes,
      totalCoverage: totalUsedBytes / totalBytes * 100,
      coverages: assets
    };
  });

  return function (_x3) {
    return _ref2.apply(this, arguments);
  };
})();

let quilted = (() => {
  var _ref = _asyncToGenerator$2(function* (url, options) {
    return yield runner(url, options);
  });

  return function quilted(_x, _x2) {
    return _ref.apply(this, arguments);
  };
})();

function _asyncToGenerator$2(fn) { return function () { var gen = fn.apply(this, arguments); return new Promise(function (resolve, reject) { function step(key, arg) { try { var info = gen[key](arg); var value = info.value; } catch (error) { reject(error); return; } if (info.done) { resolve(value); } else { return Promise.resolve(value).then(function (value) { step("next", value); }, function (err) { step("throw", err); }); } } return step("next"); }); }; }

export default quilted;
